; file: "Xlib.scm"

; A simple interface to the X-window Xlib library.

(c-declare "
#include <X11/Xlib.h>
")

; Declare a few types so that the function prototypes use the same
; type names as a C program.

(c-define-type XID unsigned-long)

(c-define-type Window XID)
(c-define-type Drawable XID)
(c-define-type Font XID)
(c-define-type Pixmap XID)
(c-define-type Cursor XID)
(c-define-type Colormap XID)
(c-define-type GContext XID)
(c-define-type KeySym XID)

(c-define-type Bool int)
(c-define-type Status int)
(c-define-type GC (pointer (struct "_XGC")))
(c-define-type Display "Display")
(c-define-type Display* (pointer Display))
(c-define-type Screen "Screen")
(c-define-type Screen* (pointer Screen))
(c-define-type XGCValues "XGCValues")
(c-define-type XGCValues* (pointer XGCValues))
(c-define-type XFontStruct "XFontStruct")
(c-define-type XFontStruct* (pointer XFontStruct))
(c-define-type XColor "XColor")
(c-define-type XColor* (pointer XColor))

(c-define-type char* char-string)

; Function prototypes for a minimal subset of Xlib functions.  The
; functions have the same name in Scheme and C.

(define XOpenDisplay
  (c-lambda (char*)        ; display_name
            Display*
            "XOpenDisplay"))

(define XCloseDisplay
  (c-lambda (Display*)     ; display
            int
            "XCloseDisplay"))

(define XDefaultScreen
  (c-lambda (Display*)     ; display
            int
            "XDefaultScreen"))

(define XScreenOfDisplay
  (c-lambda (Display*      ; display
             int)          ; screen_number
            Screen*
            "XScreenOfDisplay"))

(define XDefaultColormapOfScreen
  (c-lambda (Screen*)      ; screen
            Colormap
            "XDefaultColormapOfScreen"))

(define XClearWindow
  (c-lambda (Display*       ; display
             Window)        ; w
            int
            "XClearWindow"))

(define XRootWindow
  (c-lambda (Display*       ; display
             int)           ; screen_number
            Window
            "XRootWindow"))

(define XBlackPixel
  (c-lambda (Display*       ; display
             int)           ; screen_number
            unsigned-long
            "XBlackPixel"))

(define XWhitePixel
  (c-lambda (Display*       ; display
             int)           ; screen_number
            unsigned-long
            "XWhitePixel"))

(define XCreateSimpleWindow
  (c-lambda (Display*       ; display
             Window         ; parent
             int            ; x
             int            ; y
             unsigned-int   ; width
             unsigned-int   ; height
             unsigned-int   ; border_width
             unsigned-long  ; border
             unsigned-long) ; backgound
            Window
            "XCreateSimpleWindow"))

(define XMapWindow
  (c-lambda (Display*       ; display
             Window)        ; w
            int
            "XMapWindow"))

(define XFlush
  (c-lambda (Display*)      ; display
            int
            "XFlush"))

(define XCreateGC
  (c-lambda (Display*       ; display
             Drawable       ; d
             unsigned-long  ; valuemask
             XGCValues*)    ; values
            GC
            "XCreateGC"))

(define XFillRectangle
  (c-lambda (Display*      ; display
             Drawable      ; d
             GC            ; gc
             int           ; x
             int           ; y
             unsigned-int  ; width
             unsigned-int) ; height
            int
            "XFillRectangle"))

(define XFillArc
  (c-lambda (Display*      ; display
             Drawable      ; d
             GC            ; gc
             int           ; x
             int           ; y
             unsigned-int  ; width
             unsigned-int  ; height
             int           ; angle1
             int)          ; angle2
            int
            "XFillArc"))

(define XDrawString
  (c-lambda (Display*      ; display
             Drawable      ; d
             GC            ; gc
             int           ; x
             int           ; y
             char*         ; string
             int)          ; length
            int
            "XDrawString"))

(define XTextWidth
  (c-lambda (XFontStruct*  ; font_struct
             char*         ; string
             int)          ; count
            int
            "XTextWidth"))

(define XParseColor
  (c-lambda (Display*      ; display
             Colormap      ; colormap
             char*         ; spec
             XColor*)      ; exact_def_return
            Status
            "XParseColor"))

(define XAllocColor
  (c-lambda (Display*      ; display
             Colormap      ; colormap
             XColor*)      ; screen_in_out
            Status
            "XAllocColor"))

(define (make-XColor-box)
  ((c-lambda ()
             XColor*
             "___result = ___EXT(___alloc_mem) (sizeof (XColor));")))

(define XColor-pixel
  (c-lambda (XColor*)       ; XColor box
             unsigned-long
            "___result = ___arg1->pixel;"))

(define XColor-pixel-set!
  (c-lambda (XColor*        ; XColor box
             unsigned-long) ; intensity
            void
            "___arg1->pixel = ___arg2;"))

(define XColor-red
  (c-lambda (XColor*)       ; XColor box
             unsigned-short
            "___result = ___arg1->red;"))

(define XColor-red-set!
  (c-lambda (XColor*        ; XColor box
             unsigned-short); intensity
            void
            "___arg1->red = ___arg2;"))

(define XColor-green
  (c-lambda (XColor*)       ; XColor box
             unsigned-short
            "___result = ___arg1->green;"))

(define XColor-green-set!
  (c-lambda (XColor*        ; XColor box
             unsigned-short); intensity
            void
            "___arg1->green = ___arg2;"))

(define XColor-blue
  (c-lambda (XColor*)       ; XColor box
             unsigned-short
            "___result = ___arg1->blue;"))

(define XColor-blue-set!
  (c-lambda (XColor*        ; XColor box
             unsigned-short); intensity
            void
            "___arg1->blue = ___arg2;"))

(define (make-XGCValues-box)
  ((c-lambda ()
             XGCValues*
             "___result = ___EXT(___alloc_mem) (sizeof (XGCValues));")))

(define XGCValues-foreground
  (c-lambda (XGCValues*)    ; XGCValues box
            unsigned-long
            "return ___arg1->foreground;"))

(define XGCValues-foreground-set!
  (c-lambda (XGCValues*     ; XGCValues box
             unsigned-long) ; pixel index
            void
            "___arg1->foreground = ___arg2;"))

(define XGCValues-background
  (c-lambda (XGCValues*)    ; XGCValues box
            unsigned-long
            "return ___arg1->background;"))

(define XGCValues-background-set!
  (c-lambda (XGCValues*     ; XGCValues box
             unsigned-long) ; pixel index
            void
            "___arg1->background = ___arg2;"))

(define XGCValues-font
  (c-lambda (XGCValues*)    ; XGCValues box
            Font
            "return ___arg1->font;"))

(define XGCValues-font-set!
  (c-lambda (XGCValues*     ; XGCValues box
             Font)          ; font_ID
            void
            "___arg1->font = ___arg2;"))

(define GCFunction
  ((c-lambda () unsigned-long "___result = GCFunction;")))

(define GCPlaneMask
  ((c-lambda () unsigned-long "___result = GCPlaneMask;")))

(define GCForeground
  ((c-lambda () unsigned-long "___result = GCForeground;")))

(define GCBackground
  ((c-lambda () unsigned-long "___result = GCBackground;")))

(define GCLineWidth
  ((c-lambda () unsigned-long "___result = GCLineWidth;")))

(define GCLineStyle
  ((c-lambda () unsigned-long "___result = GCLineStyle;")))

(define GCCapStyle
  ((c-lambda () unsigned-long "___result = GCCapStyle;")))

(define GCJoinStyle
  ((c-lambda () unsigned-long "___result = GCJoinStyle;")))

(define GCFillStyle
  ((c-lambda () unsigned-long "___result = GCFillStyle;")))

(define GCFillRule
  ((c-lambda () unsigned-long "___result = GCFillRule;")))

(define GCTile
  ((c-lambda () unsigned-long "___result = GCTile;")))

(define GCStipple
  ((c-lambda () unsigned-long "___result = GCStipple;")))

(define GCTileStipXOrigin
  ((c-lambda () unsigned-long "___result = GCTileStipXOrigin;")))

(define GCTileStipYOrigin
  ((c-lambda () unsigned-long "___result = GCTileStipYOrigin;")))

(define GCFont
  ((c-lambda () unsigned-long "___result = GCFont;")))

(define GCSubwindowMode
  ((c-lambda () unsigned-long "___result = GCSubwindowMode;")))

(define GCGraphicsExposures
  ((c-lambda () unsigned-long "___result = GCGraphicsExposures;")))

(define GCClipXOrigin
  ((c-lambda () unsigned-long "___result = GCClipXOrigin;")))

(define GCClipYOrigin
  ((c-lambda () unsigned-long "___result = GCClipYOrigin;")))

(define GCClipMask
  ((c-lambda () unsigned-long "___result = GCClipMask;")))

(define GCDashOffset
  ((c-lambda () unsigned-long "___result = GCDashOffset;")))

(define GCDashList
  ((c-lambda () unsigned-long "___result = GCDashList;")))

(define GCArcMode
  ((c-lambda () unsigned-long "___result = GCArcMode;")))

(define XChangeGC
  (c-lambda (Display*       ; display
             GC             ; gc
             unsigned-long  ; valuemask
             XGCValues*)    ; values
            int
            "XChangeGC"))

(define XGetGCValues
  (c-lambda (Display*       ; display
             GC             ; gc
             unsigned-long  ; valuemask
             XGCValues*)    ; values_return
            int
            "XGetGCValues"))

(define XQueryFont
  (c-lambda (Display*       ; display
             Font)          ; font_ID
            XFontStruct*
            "XQueryFont"))

(define XLoadFont
  (c-lambda (Display*       ; display
             char*)         ; name
            Font
            "XLoadFont"))

(define XLoadQueryFont
  (c-lambda (Display*       ; display
             char*)         ; name
            XFontStruct*
            "XLoadQueryFont"))

(define XFontStruct-fid
  (c-lambda (XFontStruct*)  ; font_struct
            Font
            "___result = ___arg1->fid;"))

(define XFontStruct-ascent
  (c-lambda (XFontStruct*)  ; font_struct
            Font
            "___result = ___arg1->ascent;"))

(define XFontStruct-descent
  (c-lambda (XFontStruct*)  ; font_struct
            Font
            "___result = ___arg1->descent;"))
