;==============================================================================

; file: "_gsc.scm"

; Copyright (C) 1994-2003 by Marc Feeley, All Rights Reserved.

(include "generic.scm")

;------------------------------------------------------------------------------

(set! make-global-environment ; import runtime macros into compilation env
  (lambda ()

    (define (extract-macros cte)
      (if (##cte-top? cte)
        (env-frame #f '())
        (let ((parent-cte (##cte-parent-cte cte)))
          (if (##cte-macro? cte)
            (env-macro (extract-macros parent-cte)
                       (##cte-macro-name cte)
                       (##cte-macro-def cte))
            (extract-macros parent-cte)))))

    (extract-macros (##cte-top-cte ##interaction-cte))))

(define (compile-file-to-c
         filename
         #!optional
         (options (macro-absent-obj))
         (output (macro-absent-obj)))
  (macro-force-vars (filename)
    (macro-check-string
     filename
     1
     (compile-file-to-c filename options output)
     (let ((opts
            (if (##eq? options (macro-absent-obj))
              '()
              (macro-force-vars (options)
                options))))
       (if (##eq? output (macro-absent-obj))
         (c#cf filename #f opts #f)
         (macro-force-vars (output)
           (macro-check-string
            output
            3
            (compile-file-to-c filename options output)
            (c#cf filename #f opts output))))))))

(define (compile-file
         filename
         #!optional
         (options (macro-absent-obj))
         (cc-options (macro-absent-obj))
         (ld-options (macro-absent-obj)))
  (macro-force-vars (filename)
    (macro-check-string filename 1 (compile-file filename options)
      (let* ((opts
              (if (##eq? options (macro-absent-obj))
                '()
                (macro-force-vars (options)
                  options)))
             (cc-opts
              (if (##eq? cc-options (macro-absent-obj))
                ""
                (macro-force-vars (cc-options)
                  cc-options)))
             (ld-opts
              (if (##eq? ld-options (macro-absent-obj))
                ""
                (macro-force-vars (ld-options)
                  ld-options)))
             (x
              (c#cf filename #f opts #f)))
        (if x
          (let ((root
                 (##path-strip-directory
                  (##path-strip-extension filename))))
            (let loop ((version 1))
              (let ((root-obj
                     (##string-append
                      root
                      ".o"
                      (##number->string version 10))))
                (if (##file-exists? root-obj)
                  (loop (##fixnum.+ version 1))
                  (let* ((gambcdir
                          (##path-expand "~~"))
                         (gambcdir-bin
                          (parameterize
                           ((##current-directory
                             (##path-expand "bin" gambcdir)))
                           (##current-directory)))
                         (exit-status
                          (##shell-command
                           (##string-append
                            gambcdir-bin
                            "gsc-cc-o "
                            gambcdir
                            " "
                            root-obj
                            " "
                            root
                            ".c"
                            " "
                            cc-opts
                            " "
                            ld-opts))))
                    (if (##fixnum.= exit-status 0)
                      #t
                      '(##runtime-error;;;;;;;;;;;;;;;;;;
                        msg
                        'compile-file
                        (##list filename
                                options
                                cc-options
                                ld-options))))))))
          #f)))))

(define (link-incremental
         modules
         #!optional
         (output (macro-absent-obj))
         (base (macro-absent-obj)))
  (if (##not (##pair? modules))
    (macro-check-string modules 1 (link-incremental modules output base);;;;;;;;;;;;
      #f)
    (let loop ((lst modules) (rev-mods '()))
      (macro-force-vars (lst)
        (if (##pair? lst)
          (let ((s (##car lst)))
            (macro-force-vars (s)
              (macro-check-string-list
               s
               1
               (link-incremental modules output base)
               (loop (##cdr lst) (##cons s rev-mods)))))
          (let* ((baselib
                  (if (##eq? base (macro-absent-obj))
                    (let ((gambcdir-lib
                           (parameterize
                            ((##current-directory
                              (##path-expand "lib" (##path-expand "~~"))))
                            (##current-directory))))
                      (##string-append gambcdir-lib "_gambc"))
                    (macro-force-vars (base)
                      base)))
                 (baselib-and-mods
                  (##cons baselib (##reverse rev-mods))))
            (macro-check-string
             baselib
             3
             (link-incremental modules output base)
             (if (##eq? output (macro-absent-obj))
               (c#targ-linker #t
                              baselib-and-mods
                              #f
                              (##string-append (##car rev-mods) "_"))
               (macro-force-vars (output)
                 (macro-check-string
                  output
                  2
                  (link-incremental modules output base)
                  (c#targ-linker #t
                                 baselib-and-mods
                                 output
                                 #f)))))))))))

(define (link-flat
         modules
         #!optional
         (output (macro-absent-obj)))
  (if (##not (##pair? modules))
    (macro-check-string modules 1 (link-flat modules output);;;;;;;;;;;;;
      #f)
    (let loop ((lst modules) (rev-mods '()))
      (macro-force-vars (lst)
        (if (##pair? lst)
          (let ((s (##car lst)))
            (macro-force-vars (s)
              (macro-check-string-list
               s
               1
               (link-flat modules output)
               (loop (##cdr lst) (##cons s rev-mods)))))
          (let ((mods (##reverse rev-mods)))
            (if (##eq? output (macro-absent-obj))
              (c#targ-linker #f
                             mods
                             #f
                             (##string-append (##car rev-mods) "_"))
              (macro-force-vars (output)
                (macro-check-string
                 output
                 2
                 (link-flat modules output)
                 (c#targ-linker #f
                                mods
                                output
                                #f))))))))))

;------------------------------------------------------------------------------

#|
(##define-macro (macro-extension-file)
  "gambcext")

(##define-macro (macro-initialization-file)
  "gambcini")

(define (##main)

  (##define-macro (gsi-or x) x)

  (define (process-initialization-file)

    (define (try filename)
      (##string?
       (##load filename
               (lambda (script-line script-path) #f)
               #f
               #f)))

    (or (try (macro-initialization-file))
        (let ((homedir (##path-expand "~")))
          (try (##string-append homedir (macro-initialization-file))))))

  (define (read-source-from-string str)
    (let* ((port
            (##open-input-string str))
           (x
            (##read-all-as-a-begin-expr-from-port
             port
             ##main-readtable
             ##wrap-datum
             ##unwrap-datum
             (macro-readtable-start-syntax ##main-readtable)
             #f)))
      (##vector-ref x 1)))

  (define (interpreter-interactive-or-batch-mode arguments)
    (let loop ((lst arguments)
               (batch-mode? #f))
      (if (##pair? lst)
        (let* ((x (##car lst))
               (rest (##cdr lst)))
          (if (and (##fixnum.= 2 (##string-length x))
                   (##char=? (##string-ref x 0) #\-)
                   (##char=? (##string-ref x 1) #\e))
            (if (##pair? rest)
              (let* ((src (read-source-from-string (##car rest)))
                     (rest (##cdr rest)))
                (##eval-top src ##interaction-cte)
                (loop rest
                      batch-mode?))
              (begin
                (##repl
                 (lambda (first output-port)
                   (##write-string
                    "*** WARNING -- Missing argument for option \"-e\"\n"
                    output-port)
                   #t))
                (loop rest
                      batch-mode?)))
            (let* ((filename
                    x)
                   (loaded-script?
                    #f)
                   (script-callback
                    (lambda (script-line script-path)
                      (if script-line
                        (begin
                          (set! loaded-script?
                                (let ((language-and-tail
                                       (##extract-language-and-tail
                                        script-line)))
                                  (if language-and-tail
                                    (let ((language (##car language-and-tail)))
                                      (if (macro-language-srfi-22? language)
                                        #t; process script according to SRFI-22
                                        '()))
                                    '())))
                          (set! ##processed-command-line
                                (##cons script-path rest)))))))
              (##load filename
                      script-callback
                      #t
                      #t)
              (if loaded-script?
                (let ((status
                       (##eval
                        (if (##null? loaded-script?)
                          '(##apply main (##cdr ##processed-command-line))
                          '(main (##cdr ##processed-command-line))))))
                  (##exit status));;;;;;;;;;;;;;;
                (loop rest
                      #t)))))
        (if batch-mode?
          (##exit 0);;;;;;;;;;;;;;;;;;;;;;;;;;;
          (interactive-mode)))))

  (define (compiler-batch-mode options arguments)

    (define (scheme-file? file)
      (let ((ext (##path-extension file)))
        (or (##assoc ext ##scheme-file-extensions)
            (##equal? ext "" #f))))

    (define (c-file? file)
      (##equal? (##path-extension file) ".c" #f))

    (let* ((opts (##map ##car options))
           (sym-opts (##map ##string->symbol opts)))

      (let loop1 ((lst arguments)
                  (nb-scheme-files 0))
        (if (##not (##null? lst))

          (let ((arg (##car lst)))
            (loop1 (##cdr lst)
                   (if (scheme-file? arg)
                     (##fixnum.+ nb-scheme-files 1)
                     nb-scheme-files)))

          (let* ((gen-c?
                  (##member "c" opts))
                 (gen-dynamic?
                  (##member "dynamic" opts))
                 (link?
                  (##not (or gen-c? gen-dynamic?)))
                 (output
                  (let ((x (##assoc "o" options)))
                    (cond ((##not x)
                           #f)
                          (gen-dynamic?
                           (##repl
                            (lambda (first output-port)
                              (##write-string
                               "*** WARNING -- Dynamic object file: \"o\" option ignored\n"
                               output-port)
                              #t))
                           #f)
                          ((and (##not link?) (##fixnum.< 1 nb-scheme-files))
                           (##repl
                            (lambda (first output-port)
                              (##write-string
                               "*** WARNING -- Multiple output files: \"o\" option ignored\n"
                               output-port)
                              #t))
                           #f)
                          (else
                           (##cdr x)))))
                 (pre
                  (##assoc "prelude" options))
                 (post
                  (##assoc "postlude" options))
                 (cc-options
                  (let ((x (##assoc "cc-options" options)))
                    (if x
                      (##cdr x)
                      "")))
                 (ld-options
                  (let ((x (##assoc "ld-options" options)))
                    (if x
                      (##cdr x)
                      ""))))

            (if (or pre post)
              (set! wrap-program
                (lambda (program)
                  (##cons 'begin
                          (##append
                            (if pre
                              (let ((pre-src
                                     (read-source-from-string
                                      (##cdr pre))))
                                (##list (##desourcify pre-src)))
                              '())
                            (##cons program
                                    (if post
                                      (let ((post-src
                                             (read-source-from-string
                                              (##cdr post))))
                                        (##list (##desourcify post-src)))
                                      '())))))))

            (let loop2 ((lst arguments)
                        (rev-roots '()))
              (if (##not (##null? lst))

                (let* ((file (##car lst))
                       (root (##path-strip-extension file)))
                  (cond ((scheme-file? file)
                         (if (##fixnum.< 1 nb-scheme-files)
                           (##repl
                            (lambda (first output-port)
                              (##write-string file output-port)
                              (##write-string ":\n" output-port)
                              #t)))
                         (if gen-dynamic?
                           (if (##not (compile-file
                                       file
                                       sym-opts
                                       cc-options
                                       ld-options))
                             (##exit 1));;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
                           (let ((out (or (and (##not link?) output)
                                          (macro-absent-obj))))
                             (if (##not (compile-file-to-c
                                         file
                                         sym-opts
                                         out))
                               (##exit 1))));;;;;;;;;;;;;;;;;;;;;;;;;;;;
                       (loop2 (##cdr lst)
                              (##cons root rev-roots)))
                      ((c-file? file)
                       (loop2 (##cdr lst)
                              (##cons root rev-roots)))
                      (else
                       (##repl
                        (lambda (first output-port)
                          (##write-string file output-port)
                          (##write-string ": ignored (file not recognized)\n" output-port)
                          #t))
                       (loop2 (##cdr lst)
                              rev-roots))))

              (let* ((flat?
                      (##member "flat" opts))
                     (base
                      (let ((x (##assoc "l" options)))
                        (cond ((##not x)
                               #f)
                              ((or (##not link?) flat?)
                               (##repl
                                (lambda (first output-port)
                                  (##write-string
                                   "*** WARNING -- No incremental link: \"l\" option ignored\n"
                                   output-port)
                                  #t))
                               #f)
                              (else
                               (##cdr x))))))

                (if link?
                  (if (##not (##null? rev-roots))
                    (let ((roots (##reverse rev-roots)))
                      (if flat?
                        (link-flat
                         roots
                         (or output (macro-absent-obj)))
                        (link-incremental
                         roots
                         (or output (macro-absent-obj))
                         (or base (macro-absent-obj))))))
                  (if flat?
                    (##repl
                     (lambda (first output-port)
                       (##write-string
                        "*** WARNING -- \"c\" or \"dynamic\" option was specified: \"flat\" option ignored\n"
                        output-port)
                       #t))))))))))))

  (define (interactive-mode)
    (##set-debug-settings!
     (macro-debug-settings-error-mask)
     (macro-debug-settings-error-repl));;;;;;;;;;;;;;
    (##repl
     (lambda (first output-port)
       (if (##tty? output-port)
         (##tty-text-attributes-set! output-port 136 360));**************zzzzzzzzzzzzz
       (##write-string "Gambit Version " output-port)
       (##write-string (##system-version-string) output-port)
       (if (##tty? output-port)
         (##tty-text-attributes-set! output-port 136 392));**************zzzzzzzzzzzzz
       (##newline output-port)
       (##newline output-port)
       #f)))

  (define (##split-command-line options-with-arguments cont)

    (define (option? arg)
      (and (##fixnum.< 1 (##string-length arg))
           (##char=? (##string-ref arg 0) #\-)
           (##not (and (##fixnum.= 2 (##string-length arg))
                       (##char=? (##string-ref arg 1) #\e)))))

    (define (convert-option arg)
      (##substring arg 1 (##string-length arg)))

    (let loop1 ((arguments (##cdr ##processed-command-line))
                (rev-options '()))
      (if (and (##pair? arguments)
               (option? (##car arguments)))

        (let ((opt (convert-option (##car arguments)))
              (rest (##cdr arguments)))
          (if (##member opt options-with-arguments)
            (if (##pair? rest)
              (loop1 (##cdr rest)
                     (##cons (##cons opt (##car rest)) rev-options))
              (begin
                (##repl
                 (lambda (first output-port)
                   (##write-string
                    "*** WARNING -- Missing argument for option \""
                    output-port)
                   (##write-string opt output-port)
                   (##write-string "\"\n" output-port)
                   #t))
                (loop1 rest rev-options)))
            (loop1 rest
                   (##cons (##cons opt #f) rev-options))))

        (cont (##reverse rev-options) arguments))))

  (define (split-options lst options)
    (if (##null? lst)
      '(())
      (let* ((x (##car lst))
             (opt (##car x)))
        (if (##member opt options)
          (let ((y (split-options (##cdr lst) options)))
            (##cons (##cons x (##car y)) (##cdr y)))
          (##cons '() lst)))))

  (let ((gambcdir (##path-expand "~~")))
    (##load (##string-append gambcdir (macro-extension-file))
            (lambda (script-line script-path) #f)
            #f
            #f))

  (let ((language-and-tail
         (##extract-language-and-tail (##car ##processed-command-line))))
    (if language-and-tail
        (let ((language (##car language-and-tail)))
          (##readtable-setup-for-language! ##main-readtable language))))

  (##split-command-line
    '("o" "l" "prelude" "postlude" "cc-options" "ld-options")
    (lambda (raw-options arguments)
      (let* ((options1
              (split-options raw-options '("f" "i")))
             (main-options
              (##car options1))
             (skip-initialization-file?
              (##assoc "f" main-options))
             (force-interpreter?
              (##assoc "i" main-options))
             (other-options
              (if (gsi-or force-interpreter?)
                (split-options
                 (##cdr options1)
                 '("e"))
                (split-options
                 (##cdr options1)
                 '("c" "flat" "o" "l" "prelude" "postlude"
                   "dynamic" "cc-options" "ld-options"
                   "warnings" "verbose" "report" "expansion" "gvm"
                   "check" "force" "debug" "track-scheme"))))
             (known-options
              (##car other-options))
             (unknown-options
              (##cdr other-options)))
        (if (##not (##null? unknown-options))
          (##repl
           (lambda (first output-port)
             (##write-string "*** WARNING -- Unprocessed options: "
                             output-port)
             (##write (##map ##car unknown-options) output-port)
             (##newline output-port))))
        (if (##not skip-initialization-file?)
          (process-initialization-file))

        (if (##null? arguments)
          (interactive-mode)
          (if (gsi-or force-interpreter?)
            (interpreter-interactive-or-batch-mode arguments)
            (compiler-batch-mode known-options arguments)))))))

(define-prim (main . arguments) ; predefine main procedure so scripts don't have to
  0)

(##namespace (""))

(##main)
|#

;------------------------------------------------------------------------------

(##define-macro (macro-extension-file)
  "gambcext")

(##define-macro (macro-initialization-file)
  "gambcini")

(##define-macro (interpreter-or x)
  x)

(define (##main)

  (define (process-initialization-file)

    (define (try filename)
      (##string?
       (##load filename
               (lambda (script-line script-path) #f)
               #f
               #f)))

    (or (try (macro-initialization-file))
        (let ((homedir (##path-expand "~")))
          (try (##string-append homedir (macro-initialization-file))))))

  (define (read-source-from-string str)
    (let* ((port
            (##open-input-string str))
           (x
            (##read-all-as-a-begin-expr-from-port
             port
             ##main-readtable
             ##wrap-datum
             ##unwrap-datum
             (macro-readtable-start-syntax ##main-readtable)
             #f)))
      (##vector-ref x 1)))

  (define (interpreter-interactive-or-batch-mode arguments)
    (let loop ((lst arguments)
               (batch-mode? #f))
      (if (##pair? lst)
        (let ((file
               (##car lst))
              (rest
               (##cdr lst)))
          (if (option? file)
            (let ((option-name (convert-option file)))
              (cond ((##string=? option-name "")
                     (pp 123);;;;;;;;;;;;;;;;;;;;;;
                     (loop rest
                           batch-mode?))
                    ((##string=? option-name "e")
                     (if (##pair? rest)
                       (let ((src (read-source-from-string (##car rest))))
                         (##eval-top src ##interaction-cte)
                         (loop (##cdr rest)
                               #t))
                       (begin
                         (warn-missing-argument-for-option "e")
                         (loop rest
                               #t))))
                    (else
                     (warn-unknown-option option-name)
                     (loop rest
                           batch-mode?))))
            (let* ((loaded-script?
                    #f)
                   (script-callback
                    (lambda (script-line script-path)
                      (if script-line
                        (begin
                          (set! loaded-script?
                            (let ((language-and-tail
                                   (##extract-language-and-tail
                                    script-line)))
                              (if language-and-tail
                                (let ((language (##car language-and-tail)))
                                  (if (macro-language-srfi-22? language)
                                    #t; process script according to SRFI-22
                                    '()))
                                '())))
                          (set! ##processed-command-line
                            (##cons script-path rest)))))))

              (##load file
                      script-callback
                      #t
                      #t)

              (if loaded-script?
                (let ((status
                       (##eval
                        (if (##null? loaded-script?)
                          '(##apply main (##cdr ##processed-command-line))
                          '(main (##cdr ##processed-command-line))))))
                  (##exit status));;;;;;;;;;;;;;;
                (loop rest
                      #t)))))
        (if batch-mode?
          (##exit 0);;;;;;;;;;;;;;;;;;;;;;;;;;;
          (interactive-mode)))))

  (define (compiler-batch-mode options arguments)

    (define (scheme-file? file)
      (let ((ext (##path-extension file)))
        (or (##assoc ext ##scheme-file-extensions)
            (##string=? ext ""))))

    (define (c-file? file)
      (##string=? (##path-extension file) ".c"))

    (let* ((opts (##map ##car options))
           (sym-opts (##map ##string->symbol opts)))

      (let loop1 ((lst arguments)
                  (nb-scheme-files 0))
        (if (##pair? lst)

          (let ((file (##car lst))
                (rest (##cdr lst)))
            (cond ((option? file)
                   (let ((option-name (convert-option file)))
                     (warn-unknown-option option-name)
                     (loop1 rest
                            (##fixnum.+ nb-scheme-files 1))))
                  ((scheme-file? file)
                   (loop1 rest
                          (##fixnum.+ nb-scheme-files 1)))
                  (else
                   (loop1 rest
                          nb-scheme-files))))

          (let* ((gen-c?
                  (##member "c" opts))
                 (gen-dynamic?
                  (##member "dynamic" opts))
                 (link?
                  (##not (or gen-c? gen-dynamic?)))
                 (output
                  (let ((x (##assoc "o" options)))
                    (cond ((##not x)
                           #f)
                          (gen-dynamic?
                           (warn-dynamic-object-file-and-o-option)
                           #f)
                          ((and (##not link?) (##fixnum.< 1 nb-scheme-files))
                           (warn-multiple-output-files-and-o-option)
                           #f)
                          (else
                           (##cdr x)))))
                 (pre
                  (##assoc "prelude" options))
                 (post
                  (##assoc "postlude" options))
                 (cc-options
                  (let ((x (##assoc "cc-options" options)))
                    (if x
                      (##cdr x)
                      "")))
                 (ld-options
                  (let ((x (##assoc "ld-options" options)))
                    (if x
                      (##cdr x)
                      ""))))

            (if (or pre post)
              (set! wrap-program
                (lambda (program)
                  (##cons 'begin
                          (##append
                            (if pre
                              (let ((pre-src
                                     (read-source-from-string
                                      (##cdr pre))))
                                (##list (##desourcify pre-src)))
                              '())
                            (##cons program
                                    (if post
                                      (let ((post-src
                                             (read-source-from-string
                                              (##cdr post))))
                                        (##list (##desourcify post-src)))
                                      '())))))))

            (let loop2 ((lst arguments)
                        (rev-roots '()))
              (if (##pair? lst)

                (let ((file (##car lst))
                      (rest (##cdr lst)))
                  (if (option? file)
                    (loop2 rest
                           rev-roots)
                    (let ((root (##path-strip-extension file)))
                      (cond ((scheme-file? file)
                             (if (##fixnum.< 1 nb-scheme-files)
                               (##repl
                                (lambda (first output-port)
                                  (##write-string file output-port)
                                  (##write-string ":\n" output-port)
                                  #t)))
                             (if gen-dynamic?
                               (if (##not (compile-file
                                           file
                                           sym-opts
                                           cc-options
                                           ld-options))
                                 (##exit 1));;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
                               (let ((out (or (and (##not link?) output)
                                              (macro-absent-obj))))
                                 (if (##not (compile-file-to-c
                                             file
                                             sym-opts
                                             out))
                                   (##exit 1))));;;;;;;;;;;;;;;;;;;;;;;;;;;;
                             (loop2 rest
                                    (##cons root rev-roots)))

                            ((c-file? file)
                             (loop2 rest
                                    (##cons root rev-roots)))

                            (else
                             (warn-file-not-recognized file)
                             (loop2 rest
                                    rev-roots))))))

                (let* ((flat?
                        (##member "flat" opts))
                       (base
                        (let ((x (##assoc "l" options)))
                          (cond ((##not x)
                                 #f)
                                ((or (##not link?) flat?)
                                 (warn-no-incremental-link)
                                 #f)
                                (else
                                 (##cdr x))))))

                  (if link?
                    (if (##not (##null? rev-roots))
                      (let ((roots (##reverse rev-roots)))
                        (if flat?
                          (link-flat
                           roots
                           (or output (macro-absent-obj)))
                          (link-incremental
                           roots
                           (or output (macro-absent-obj))
                           (or base (macro-absent-obj))))))
                    (if flat?
                      (warn-c-or-dynamic-and-flat)))))))))))

  (define (interactive-mode)
    (##set-debug-settings!
     (macro-debug-settings-error-mask)
     (macro-debug-settings-error-repl));;;;;;;;;;;;;;
    (##repl
     (lambda (first output-port)
       (if (##tty? output-port)
         (##tty-text-attributes-set! output-port 136 360));**************zzzzzzzzzzzzz
       (##write-string "Gambit Version " output-port)
       (##write-string (##system-version-string) output-port)
       (if (##tty? output-port)
         (##tty-text-attributes-set! output-port 136 392));**************zzzzzzzzzzzzz
       (##newline output-port)
       (##newline output-port)
       #f)))

  (define (warn-missing-argument-for-option opt)
    (##repl
     (lambda (first output-port)
       (##write-string
        "*** WARNING -- Missing argument for option \""
        output-port)
       (##write-string opt output-port)
       (##write-string "\"\n" output-port)
       #t)))

  (define (warn-unknown-option opt)
    (##repl
     (lambda (first output-port)
       (##write-string
        "*** WARNING -- Unknown or improperly placed option: "
        output-port)
       (##write opt output-port)
       (##newline output-port)
       #t)))

  (define (warn-dynamic-object-file-and-o-option)
    (##repl
     (lambda (first output-port)
       (##write-string
        "*** WARNING -- Dynamic object file: \"o\" option ignored\n"
        output-port)
       #t)))

  (define (warn-multiple-output-files-and-o-option)
    (##repl
     (lambda (first output-port)
       (##write-string
        "*** WARNING -- Multiple output files: \"o\" option ignored\n"
        output-port)
       #t)))

  (define (warn-file-not-recognized file)
    (##repl
     (lambda (first output-port)
       (##write-string file output-port)
       (##write-string ": ignored (file not recognized)\n" output-port)
       #t)))

  (define (warn-no-incremental-link)
    (##repl
     (lambda (first output-port)
       (##write-string
        "*** WARNING -- No incremental link: \"l\" option ignored\n"
        output-port)
       #t)))

  (define (warn-c-or-dynamic-and-flat)
    (##repl
     (lambda (first output-port)
       (##write-string
        "*** WARNING -- \"c\" or \"dynamic\" option was specified: \"flat\" option ignored\n"
        output-port)
       #t)))

  (define (option? arg)
    (and (##fixnum.< 0 (##string-length arg))
         (##char=? (##string-ref arg 0) #\-)))

  (define (convert-option arg)
    (##substring arg 1 (##string-length arg)))

  (define (split-command-line
           arguments
           options-with-no-args
           options-with-args
           cont)
    (let loop1 ((args arguments)
                (rev-options '()))
      (if (and (##pair? args)
               (option? (##car args)))

        (let ((opt (convert-option (##car args)))
              (rest (##cdr args)))
          (cond ((##member opt options-with-no-args)
                 (loop1 rest
                        (##cons (##cons opt #f) rev-options)))
                ((##member opt options-with-args)
                 (if (##pair? rest)
                   (loop1 (##cdr rest)
                          (##cons (##cons opt (##car rest)) rev-options))
                   (begin
                     (warn-missing-argument-for-option opt)
                     (loop1 rest rev-options))))
                (else
                 (cont (##reverse rev-options) args))))

        (cont (##reverse rev-options) args))))

  (let ((gambcdir (##path-expand "~~")))
    (##load (##string-append gambcdir (macro-extension-file))
            (lambda (script-line script-path) #f)
            #f
            #f))

  (let ((language-and-tail
         (##extract-language-and-tail (##car ##processed-command-line))))
    (if language-and-tail
        (let ((language (##car language-and-tail)))
          (##readtable-setup-for-language! ##main-readtable language))))

  (split-command-line
    (##cdr ##processed-command-line)
    '("f" "i")
    '()
    (lambda (main-options arguments)
      (let ((skip-initialization-file?
             (##assoc "f" main-options))
            (force-interpreter?
             (##assoc "i" main-options)))
        (split-command-line
          arguments
          (if (interpreter-or force-interpreter?)
            '()
            '("c" "flat" "dynamic"
              "warnings" "verbose" "report" "expansion" "gvm"
              "check" "force" "debug" "track-scheme"))
          (if (interpreter-or force-interpreter?)
            '()
            '("o" "l" "prelude" "postlude" "cc-options" "ld-options"))
          (lambda (known-options arguments)

            (if (##not skip-initialization-file?)
              (process-initialization-file))

            (if (or (##null? arguments)
                    (interpreter-or force-interpreter?))
              (interpreter-interactive-or-batch-mode arguments)
              (compiler-batch-mode known-options arguments))))))))

(define-prim (main . arguments) ; predefine main procedure so scripts don't have to
  0)

(##namespace (""))

(##main)
